/*
**  MATLAB-to-Mathematica MathLink Connection
**  Version 1.2
**
**  Author:			Douglas B. Stein (doug@wri.com)
**  Date:			17 January 1994
**  File:           math35.c
**  Description:    The MathLink MEX resource.
**  Platform:       Macintosh
**
**  Version History
**      1.0 - Initial port for MATLAB 3.5 on the Macintosh
**      1.1 - Modifications for MATLAB 4.0 on all platforms
**		1.2 - Modifications to support K&R compilers and MATLAB 4.1
**				(Mac and Unix - Windows to come on next version)
**				This file only supports MATLAB 3.5. Support for
**				MATLAB 4.x is implemented by the file math.c.
**
**
*/

#include <mathlink.h>
#include "cmex.h"

#ifdef applec
#include <OSUtils.h>
#include <Memory.h>
#include <Quickdraw.h>
#include <Types.h>
pascal long GetA5(void) = 0x2E8D;	/* move.l A5,(A7) */
#else
enum {false,true};
typedef unsigned char Boolean;
#ifndef NULL
#define NULL ((void *) 0)
#endif
#endif

Boolean first_time = true;
Boolean swapEOL = false;
MLINK mlp = NULL;

#define MEXARGUMENTERROR			0
#define STRINGEVAL					1
#define PUTARRAYTOMATHEMATICA		2
#define GETARRAYFROMMATHEMATICA		3
#define CLOSEMATHLINK				4
#define PROCESSEDOPENLINKARGUMENTS	5

#define LINKOPENCOMMAND				"linkopen"
#define PUTARRAYCOMMAND				"matlab2math"
#define	GETARRAYCOMMAND				"math2matlab"
#define QUITCOMMAND					"quit"
#define EXITCOMMAND					"exit"
#define SWAPEOLSTRING				"swapEOL"

#ifdef applec
#define mex_qd qd
#define MATLABMAX					1022
#endif

#ifdef __STDC__
extern void mexmain(void);
static Boolean OpenMathLink(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[]);
static void CloseMathLink(void);
static void MathLinkStringEval(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[]);
static INT  ParseMatlabArgs(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[]);
static void PutArrayToMathematica(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[]);
static void GetArrayFromMathematica(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[]);
static void MatToStr(Matrix *mat, char **str);
static void StrToMat(char *str, Matrix **mat);
static int	 WaitForReturnPacket(MLINK mlp);
static void HandleMathLinkError(MLINK mlp, Matrix **mat);
static void SwapLineTerminators(char *str, unsigned int len);
#else
extern mexmain();
static Boolean OpenMathLink();
static CloseMathLink();
static MathLinkStringEval();
static INT  ParseMatlabArgs();
static PutArrayToMathematica();
static GetArrayFromMathematica();
static MatToStr();
static StrToMat();
static int	 WaitForReturnPacket();
static HandleMathLinkError();
static SwapLineTerminators();
#endif

#ifdef __STDC__
INT user_fcn(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[])
#else
INT user_fcn(nlhs, plhs, nrhs, prhs)
INT nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
	/* Check for proper number of arguments */
	switch (ParseMatlabArgs(nlhs, plhs, nrhs, prhs)) {
		case MEXARGUMENTERROR:
			mex_error("Not valid format for math().");
			break;
		case STRINGEVAL:
			MathLinkStringEval(nlhs, plhs, nrhs, prhs);
			break;
		case PUTARRAYTOMATHEMATICA:
			PutArrayToMathematica(nlhs, plhs, nrhs, prhs);
			break;
		case GETARRAYFROMMATHEMATICA:
			GetArrayFromMathematica(nlhs, plhs, nrhs, prhs);
			break;
		case CLOSEMATHLINK:
			CloseMathLink();
			break;
	}
	
	return 1;
}

static
#ifdef __STDC__
INT ParseMatlabArgs(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[])
#else
INT ParseMatlabArgs(nlhs, plhs, nrhs, prhs)
INT nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
	int	res = MEXARGUMENTERROR;
	char *str = NULL;
	
	if (first_time) {
		if (OpenMathLink(nlhs, plhs, nrhs, prhs))
			return PROCESSEDOPENLINKARGUMENTS;
	}
	
	if (nlhs > 1)
		mex_error("math() requires zero or one output arguments.");
	else if (nrhs < 1)
		mex_error("math() requires at least one input argument.");
	else if (prhs[0]->type != TEXT)
		mex_error("The first input argument for math() must be a string.");
	else  {/* 1 or more rhs args - the first is a string */
		MatToStr(prhs[0], &str);
		if (nrhs == 1) {
			if (!strcmp(str, QUITCOMMAND) || !strcmp(str, EXITCOMMAND))
				res = CLOSEMATHLINK;
			else
				res = STRINGEVAL;
		} else {/* either matlab2math or math2matlab */
			if (!strcmp(str, PUTARRAYCOMMAND) && (nrhs == 3) &&
				(prhs[1]->type == TEXT) && (prhs[2]->type != TEXT))
				res = PUTARRAYTOMATHEMATICA;
			else if (!strcmp(str, GETARRAYCOMMAND) && (nrhs == 2) &&
				(prhs[1]->type == TEXT))
				res = GETARRAYFROMMATHEMATICA;
			else
				res = MEXARGUMENTERROR;
		}
		free(str);
	}
	
	return res;
}

static
#ifdef __STDC__
Boolean OpenMathLink(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[])
#else
Boolean OpenMathLink(nlhs, plhs, nrhs, prhs)
INT nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
	long		localA5;
	int			argc;
	char		*argv[8], *commandStr, *swapOpt;
	int			mlres;
	Boolean		processedArgs = false;
	
#ifdef applec
	InitGraf(&mex_qd.thePort);
#endif
	
#ifdef applec
	argc = 7;
	argv[0] = "MathLinkMex";
	argv[1] = "-linkname";		argv[2] = "MathKernel";
	argv[3] = "-linkmode";		argv[4] = "Launch";
	argv[5] = "-linkprotocol";	argv[6] = "PPC";
	argv[7] = "\0";
#else
	argc = 7;
	argv[0] = "MathLinkMex";
    argv[1] = "-linkname";      argv[2] = "math -mathlink";
	argv[3] = "-linkmode";		argv[4] = "Launch";
	argv[5] = "-linkprotocol";	argv[6] = "Pipes";
	argv[7] = "\0";
#endif

	swapEOL = false;
	
	MatToStr(prhs[0], &commandStr);
	/* look for argument of the form 									*/
	/* ('linkopen', 'linkname', 'linkmode', 'linkprotocol', 'swapEOL')	*/
	/* otherwise we'll use the default arguments above					*/
	if (!strcmp(commandStr, LINKOPENCOMMAND) && (nrhs == 5) &&
		(prhs[1]->type == TEXT) && (prhs[2]->type == TEXT) &&
		(prhs[3]->type == TEXT) && (prhs[4]->type == TEXT)) {
		MatToStr(prhs[1], &argv[2]);
		MatToStr(prhs[2], &argv[4]);
		MatToStr(prhs[3], &argv[6]);
		MatToStr(prhs[4], &swapOpt);
		if (!strcmp(swapOpt, SWAPEOLSTRING)) swapEOL = true;
		processedArgs = true;
	}
	
#ifdef applec
	localA5 = GetA5();
#endif
    mlp = MLOpen(argc, argv);
#ifdef applec
	SetA5(localA5);
#endif

	if (processedArgs) {
		free(argv[2]);
		free(argv[4]);
		free(argv[6]);
		free(swapOpt);
	}
	free(commandStr);
	
	if (mlp == NULL) {
		mex_error("MathLink connection unexpectedly NULL!");
	}
	
	first_time = false;
	mex_printf("Mathematica Kernel loading...\n");
	
	/* now preload some convenient definitions into the Mathematica kernel */
	MLPutFunction(mlp, "ToExpression", 1);
		MLPutFunction(mlp, "StringJoin", 5);
		MLPutString(mlp, "MathToMATLAB::badform = \"Second argument to math() not expressible as a MATLAB array.\";");
		MLPutString(mlp, "MATLABQ[x_]:=VectorQ[x,NumberQ[N[#]]&]||MatrixQ[x,NumberQ[N[#]]&];");
		MLPutString(mlp, "MathToMATLAB[x_?MATLABQ] := MATLABArray[N[x]];");
 		MLPutString(mlp, "MathToMATLAB[x_] := (Message[MathToMATLAB::badform];$Failed);");
 		MLPutString(mlp, "SetOptions[ToString, PageWidth->70];");
	MLEndPacket(mlp);
	mlres = WaitForReturnPacket(mlp);

	if (!mlres) {
		HandleMathLinkError(mlp, &plhs[0]);
	}
	MLNewPacket(mlp);
	
	return processedArgs;
}

static
#ifdef __STDC__
void CloseMathLink(void)
#else
CloseMathLink()
#endif
{
	if (!first_time) {
		MLClose(mlp);
        mlp = NULL;
		first_time = true;
		mex_printf("Mathematica Kernel quitting per your request...\n");
	}
	
	return;
}

static
#ifdef __STDC__
void MathLinkStringEval(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[])
#else
MathLinkStringEval(nlhs, plhs, nrhs, prhs)
INT nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
	int		mlres;
	char	*inStr = NULL;
	char	*outStr = NULL;

	if (mlp == NULL)
		mex_error("MathLink connection unexpectedly NULL!");
		
	MatToStr(prhs[0], &inStr);

	MLPutFunction(mlp, "ToString", 1);
		MLPutFunction(mlp, "ToExpression", 1);
			MLPutString(mlp, inStr);
	mlres = MLEndPacket(mlp);
	free(inStr);
	
	mlres = WaitForReturnPacket(mlp);

	if (mlres) {
		MLGetString(mlp, &outStr);
		StrToMat(outStr, &plhs[0]);
		MLDisownString(mlp, outStr);
	} else {
		HandleMathLinkError(mlp, &plhs[0]);
	}
	MLNewPacket(mlp);

	return;
}

static
#ifdef __STDC__
void PutArrayToMathematica(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[])
#else
PutArrayToMathematica(nlhs, plhs, nrhs, prhs)
INT nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
	int			mlres;
	char		*nameStr = NULL;
	DOUBLE		*data, *currPtr, *realPtr, *imagPtr;
	long		dims[3];
	char		*heads[3];
	long		depth;
	unsigned INT	row, col;

	/* I'd prefer to use aggregate initializers, but K&R compilers */
	/* such as that shipped standard by Sun are retarded in this */
	/* respect. */
	dims[0] = dims[1] = dims[2] = 0;
	heads[0] = "List"; heads[1] = "List"; heads[2] = "Complex";

	if (mlp == NULL)
		mex_error("MathLink connection unexpectedly NULL!");
		
	MatToStr(prhs[1], &nameStr);

	dims[0] = prhs[2]->m;
	dims[1] = prhs[2]->n;
	dims[2] = 2;
	
	if (prhs[2]->pi == NULL) {	/* real matrix */
		depth = 2;
		realPtr = prhs[2]->pr;
		data = (DOUBLE *)malloc(dims[0] * dims[1] * sizeof(DOUBLE));
		if (!data)
			mex_error("Memory allocation failed in PutArrayToMathematica!");
		for (currPtr = data, row = 0; row < dims[0]; row++)
			for (col = 0; col < dims[1]; col++)
				*(currPtr++) = realPtr[col*dims[0]+row];
	} else {	/* complex matrix */
		depth = 3;
		data = (DOUBLE *)malloc(dims[0] * dims[1] * dims[2] * sizeof(DOUBLE));
		if (!data)
			mex_error("Memory allocation failed in PutArrayToMathematica!");
		realPtr = prhs[2]->pr;
		imagPtr = prhs[2]->pi;
		for (currPtr = data, row = 0; row < dims[0]; row++)
			for (col = 0; col < dims[1]; col++) {
				*(currPtr++) = realPtr[col*dims[0]+row];
				*(currPtr++) = imagPtr[col*dims[0]+row];
			}
	}
	
	MLPutFunction(mlp, "CompoundExpression", 2);
		MLPutFunction(mlp, "Set", 2);
			MLPutSymbol(mlp, nameStr);
			MLPutDoubleArray(mlp, data, dims, heads, depth);
		free(data);
		MLPutSymbol(mlp, "Null");
	mlres = MLEndPacket(mlp);
	
	mlres = WaitForReturnPacket(mlp);

	if (mlres) {
		StrToMat(nameStr, &plhs[0]);
	} else {
		HandleMathLinkError(mlp, &plhs[0]);
	}
	MLNewPacket(mlp);
	
	free(nameStr);
	return;
}

static
#ifdef __STDC__
void GetArrayFromMathematica(INT nlhs, Matrix *plhs[], INT nrhs, Matrix *prhs[])
#else
GetArrayFromMathematica(nlhs, plhs, nrhs, prhs)
INT nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
	int			mlres;
	char		*nameStr = NULL;
	DOUBLE		*data, *currPtr, *realPtr, *imagPtr;
	long		*dims;
	char		**heads;
	long		depth, argcount;
	unsigned INT	row, col;

	if (mlp == NULL)
		mex_error("MathLink connection unexpectedly NULL!");
		
	MatToStr(prhs[1], &nameStr);

	MLPutFunction(mlp, "MathToMATLAB", 1);
		MLPutFunction(mlp, "ToExpression", 1);
			MLPutString(mlp, nameStr);
	mlres = MLEndPacket(mlp);
	
	mlres = WaitForReturnPacket(mlp);

	if ((MLGetType(mlp) == MLTKFUNC) &&
		(MLCheckFunction(mlp, "MATLABArray", &argcount))) {
		mlres = mlres && MLGetDoubleArray(mlp, &data, &dims, &heads, &depth);
		if (mlres) {
			switch (depth) {
				case 1:	/* 1D real */
					plhs[0] = create_matrix(1, dims[0], REAL);
					realPtr = plhs[0]->pr;
					for (currPtr = data, col = 0; col < dims[0]; col++)
						realPtr[col] = *(currPtr++);
					break;
				case 2:	/* 1D complex or 2D real */
					if (!strcmp(heads[1], "Complex")) {
						/* 1D complex */
						plhs[0] = create_matrix(1, dims[0], COMPLEX);
						realPtr = plhs[0]->pr;
						imagPtr = plhs[0]->pi;
						for (currPtr = data, col = 0; col < dims[0]; col++) {
							realPtr[col] = *(currPtr++);
							imagPtr[col] = *(currPtr++);
						}
					} else {
						/* 2D real */
						plhs[0] = create_matrix(dims[0], dims[1], REAL);
						realPtr = plhs[0]->pr;
						for (currPtr = data, row = 0; row < dims[0]; row++)
							for (col = 0; col < dims[1]; col++)
								realPtr[col*dims[0]+row] = *(currPtr++);
					}
					break;
				case 3: /* 2D Complex */
					if (!strcmp(heads[2], "Complex")) {
						plhs[0] = create_matrix(dims[0], dims[1], COMPLEX);
						realPtr = plhs[0]->pr;
						imagPtr = plhs[0]->pi;
						for (currPtr = data, row = 0; row < dims[0]; row++)
							for (col = 0; col < dims[1]; col++) {
								realPtr[col*dims[0]+row] = *(currPtr++);
								imagPtr[col*dims[0]+row] = *(currPtr++);
							}
					} else {
						mex_printf("MATLAB can handle only 1D or 2D arrays!\n");
					}
					break;
				default:
					mex_printf("MATLAB can handle only 1D or 2D arrays!\n");
			}
			MLDisownDoubleArray(mlp, data, dims, heads, depth);
		} else {
			HandleMathLinkError(mlp, &plhs[0]);
		}
	} else {
		StrToMat("$Failed", &plhs[0]);
	}
	MLNewPacket(mlp);
	
	free(nameStr);
	return;
}

static
#ifdef __STDC__
void MatToStr(Matrix *mat, char **str)
#else
MatToStr(mat, str)
Matrix *mat;
char **str;
#endif
{
	unsigned int	len, index;
	char			*charPtr;
	DOUBLE			*doublePtr;

	len = mat->n;
	*str = (char *)malloc((len + 1) * sizeof(char));
	for (charPtr = *str, doublePtr = mat->pr, index = 0;
		 index < len;
		 index++)	{
		*(charPtr++) = (char) (*(doublePtr++));
	}
	*charPtr = '\0';
	if (swapEOL) 
		SwapLineTerminators(*str, len);
	return;
}

static
#ifdef __STDC__
void StrToMat(char *str, Matrix **mat)
#else
StrToMat(str, mat)
char *str;
Matrix **mat;
#endif
{
	unsigned int	len, index;
	char			*charPtr;
	DOUBLE			*doublePtr;

	len = strlen(str);
	
#ifdef applec
	if (len > MATLABMAX) {
		mex_printf("MATLAB crashes for strings longer than %d - trimming to fit...\n", MATLABMAX);
		len = MATLABMAX;
	}
#endif
	
	if (swapEOL)
		SwapLineTerminators(str, len);
		
	*mat = create_matrix(1, len, REAL);
	(*mat)->type = TEXT;
	for (charPtr = str, doublePtr = (*mat)->pr, index = 0;
		index < len; index++)	{
		*(doublePtr++) = (DOUBLE) (*(charPtr++));
	}
	return;
}

static
#ifdef __STDC__
int	WaitForReturnPacket(MLINK mlp)
#else
int WaitForReturnPacket(mlp)
MLINK mlp;
#endif
{
	int mlres;
	unsigned int len;
	char *msgStr, *destStr;
	
	while ((mlres = MLNextPacket(mlp)) && (mlres != RETURNPKT)) {
		if (mlres == TEXTPKT) {
			MLGetString(mlp, &msgStr);
			if (!swapEOL)
				mex_printf("%s\n", msgStr); /* easy case */
			else { /* need to copy before mutating the string */
				len = strlen(msgStr);
                destStr = (char *)malloc((len + 1) * sizeof(char));
				if (destStr) { /* copied successfully - can mutate string */
					(void)strcpy(destStr, msgStr);
					SwapLineTerminators(destStr, len);
					mex_printf("%s\n", destStr);
					free(destStr);
				} else /* couldn't copy - should at least output string as is */
					mex_printf("%s\n", msgStr);
			}
			MLDisownString(mlp, msgStr);
		}
		MLNewPacket(mlp);
	}
	return mlres;
}

static
#ifdef __STDC__
void HandleMathLinkError(MLINK mlp, Matrix **mat)
#else
HandleMathLinkError(mlp, mat)
MLINK mlp;
Matrix **mat;
#endif
{
	mex_printf("%s\n", MLErrorMessage(mlp));
	MLClearError(mlp);
	StrToMat("$Failed", mat);
	return;
}

static
#ifdef __STDC__
void SwapLineTerminators(char *str, unsigned int len)
#else
SwapLineTerminators(str, len)
char *str;
unsigned int len;
#endif
{
	unsigned int index;
	char *charPtr;
	
	for (charPtr = str, index = 0;
		 index < len;
		 index++)	{
		 if (*charPtr == '\r')
			*charPtr = '\n';
		 else if (*charPtr == '\n')
			*charPtr = '\r';
		charPtr++;
	}
	return;
}
